/*
    This program demonstrates how a "race condition" can come up
    when two threads both manipulate a shared variable. In this
    example, the shared variable is the global variable count.
    A race condition is where the result of a piece of code
    depends on the order in which the processor executes the
    threads.
*/
#include <windows.h>
#include <stdio.h>

// global variables
int runFlag = TRUE;
unsigned int count  = 0;  // this variable is shared by thread1 and thread2
unsigned int count1 = 0;  // this variable is used only by thread1
unsigned int count2 = 0;  // this variable is used only by thread2

int main( int argc, char *argv[])
{
   DWORD WINAPI thread1(LPVOID);  //thread function prototypes
   DWORD WINAPI thread2(LPVOID);

   int threadParam1 = 0;  // thread function parameters
   int threadParam2 = 0;

   // create two threads
   printf("race1: Starting threads\n");
   CreateThread(NULL, 0, thread1, &threadParam1, 0, NULL);
   CreateThread(NULL, 0, thread2, &threadParam2, 0, NULL);

   Sleep(2000);      // let the threads run for a while
   runFlag = FALSE;  // now stop the threads
   Sleep(20);        // give the threads time to stop

   printf("\nThreads stopped.\n");
   printf("count = %d, count1 = %d, count2 = %d\n", count, count1, count2);

   return 0;
}//main


DWORD WINAPI thread1(LPVOID threadParam)
{
   unsigned int my_count;
   while(runFlag)
   {
      // create a race situation with thread2
      my_count = count++;
      if (count - my_count != 1)
         printf("Bingo1: count - my_count = %d\n", count-my_count);
      count1++;
   }
   return 0;
}//thread1


DWORD WINAPI thread2(LPVOID threadParam)
{
   unsigned int my_count;
   while(runFlag)
   {
      // create a race situation with thread1
      my_count = count++;
      if (count - my_count != 1)
         printf("Bingo2: count - my_count = %d\n", count-my_count);
      count2++;
   }
   return 0;
}//thread2
